from .webapi import EnumParam as EnumParam, APIObj as APIObj, WebAPI as WebAPI
from .oauth2 import OAuth2 as OAuth2, OAuth2User as OAuth2User
from .oauth1 import OAuth1 as OAuth1, OAuth1User as OAuth1User
from .auth import APIKey as APIKey
from .asyncy import AsyncTrans as AsyncTrans, AsyncLazy as AsyncLazy


from typing import Callable
from .asyncy import *
from .webapi import *
import functools

Endpoint = Callable[[WebAPI, T], Coroutine[Any, None, U]]

def ratelimit(bucket: str, bucket_params: list[str]) -> Callable[[Endpoint[Any, T]], Endpoint[Any, T]]:
    def decorator(func: Endpoint[Any, T]) -> Endpoint[Any, T]:
        @functools.wraps(func)
        async def wrapper(self: WebAPI, params: dict[str, Any]) -> T:
            # TODO
            return await func(self, params)
        return wrapper
    return decorator


import asyncio
from datetime import datetime, timedelta
from dataclasses import dataclass, field
import json
import secrets
from typing import Any

from .auth import Auth
from .web import Request, serve_one_request

import aiohttp, aiohttp.web

import urllib.parse

@dataclass
class OAuth2User:
    token: str
    refresh_token: str
    expires_at: datetime
    token_type: str = 'Bearer'
    source_path: str | None = None
    scopes: list[str] = field(default_factory=list)

    def __init__(self, source: str | dict[str, Any], source_path: str | None = None) -> None:
        match source:
            case str(): # file path
                with open(source, 'r') as f:
                    data = json.load(f)
                    self.__dict__ |= OAuth2User(data).__dict__
                    self.source_path = source
            case {
                'token': token,
                'refresh_token': refresh_token,
                'expires_at': expires_at_str,
                'token_type': token_type,
                'scopes': scopes
            }: #dumps
                self.token = token
                self.refresh_token = refresh_token
                self.expires_at = datetime.strptime(expires_at_str, '%Y-%m-%dT%H:%M:%S.%fZ')
                self.token_type = token_type
                self.source_path = source_path
                self.scopes = scopes
            case {
                'access_token': token,
                'refresh_token': refresh_token,
                'expires_in': expires_str,
                'token_type': token_type,
                **others
            }: # OAuth 2 grant response
                expiry = datetime.utcnow() + timedelta(seconds=int(expires_str))
                self.token = token
                self.refresh_token = refresh_token
                self.expires_at = expiry
                self.token_type = token_type
                self.source_path = source_path
                if 'scope' in others:
                    self.scopes = others['scope'].split(' ')
            case _:
                raise ValueError(F"Invalid OAuth2User source: {source}")

    def to_dict(self) -> dict[str, str]:
        return {
            'token': self.token,
            'refresh_token': self.refresh_token,
            'expires_at': self.expires_at.strftime('%Y-%m-%dT%H:%M:%S.%fZ'),
            'token_type': self.token_type,
        }

    def save(self):
        if self.source_path:
            with open(self.source_path, 'w') as f:
                json.dump(self.to_dict(), f, indent=4)
        else:
            raise ValueError("Cannot save OAuth2User without a path to save to")

@dataclass
class OAuth2(Auth):
    id: str
    secret: str

    auth_uri: str # flow step 1
    token_uri: str # flow step 2

    user: OAuth2User | None = None

    _refresh_task: asyncio.Task[None] | None = None

    def __init__(self, source: str | dict[str, Any], user: OAuth2User | None = None) -> None:
        match source:
            case str(): # file path
                with open(source, 'r') as f:
                    data = json.load(f)
                    self.__dict__ |= OAuth2(data).__dict__
                    # self.source_path = source
            case { 'id': id_, 'secret': secret, 'auth_uri': auth_uri, 'token_uri': token_uri}: # dumps
                self.id = id_
                self.secret = secret
                self.auth_uri = auth_uri
                self.token_uri = token_uri
            case _:
                raise ValueError(F"Invalid OAuth2 source: {source}")
        self.user = user

    def verify_scope(self, scope: str):
        scopes = scope.split(' ')
        for scope in scopes:
            if not self.user:
                raise ValueError("User credentials must be used when scope is specified")
            if not scope in self.user.scopes:
                raise ValueError(F"User credentials do not have the scope: {scope}")
        return True

    def get_auth_url(self, redirect_uri: str, state: str, scopes: str) -> tuple[str, str]:
        challenge = secrets.token_urlsafe(54)
        params = {
            'client_id': self.id,
            'redirect_uri': redirect_uri,
            'response_type': 'code',
            'state': state+challenge,
            'scope': scopes,
            'access_type': 'offline'
            }
        return F"{self.auth_uri}?{urllib.parse.urlencode(params)}", challenge

    async def sign_request(self, session: aiohttp.ClientSession, request: Request, do_user: bool = True) -> Request:
        if self.user is not None and do_user:
            if self._refresh_task is not None:
                await asyncio.wait_for(self._refresh_task, timeout=None)
            elif datetime.utcnow() > self.user.expires_at:
                self._refresh_task = asyncio.create_task(self.refresh(session))
                await self._refresh_task
                self._refresh_task = None
            request.headers['Authorization'] = F"{self.user.token_type} {self.user.token}"
        else:
            raise NotImplemented("OAuth request signing without user is not yet implemented.")
        return request

    async def refresh(self, session: aiohttp.ClientSession):
        if self.user is None:
            raise NotImplemented("Refresh without user is not implemented.")
        data = {
            'grant_type': 'refresh_token',
            'refresh_token': self.user.refresh_token,
            'client_id': self.id,
            'client_secret': self.secret,
        }
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}

        async with session.post(self.token_uri, data=data, headers=headers) as resp:
            if resp.status != 200:
                raise Exception(f'Refresh failed: {resp.status}')
            result = await resp.json()

        match result:
            case {
                'access_token': token,
                'expires_in': expires_str,
                'token_type': token_type,
                **_others
            }: # OAuth 2 refresh response
                expiry = datetime.utcnow() + timedelta(seconds=int(expires_str))
                self.user.token = token
                self.user.expires_at = expiry
                self.user.token_type = token_type
            case _:
                raise ValueError(F"Invalid OAuth2 refresh response: {result}")
        self.user.save()
         
    async def user_auth_flow(self, redirect_host: str, redirect_port: int, **kwargs: str):
        import webbrowser

        redirect_uri = F'http://{redirect_host}:{redirect_port}'

        session = aiohttp.ClientSession()

        scopes: str = kwargs['scopes']

        # step 1: get the user to authorize the application
        grant_link, challenge = self.get_auth_url(redirect_uri, '', scopes)

        webbrowser.open(grant_link, new=1, autoraise=True)

        # step 1 (cont.): wait for the user to be redirected with the code
        query = await serve_one_request(redirect_host, redirect_port, '<html><body>You can close this window now.</body></html>')

        if 'state' not in query:
            raise PermissionError("Redirect did not return any state parameter.")
        if not query['state'] == challenge:
            raise PermissionError("Redirect did not return the correct state parameter.")
        code = query['code']

        # step 2: exchange the code for access token
        grant_data = {
            'grant_type': 'authorization_code',
            'code': code,
            'client_id': self.id,
            'client_secret': self.secret,
            'redirect_uri': redirect_uri,
            'scope': scopes
        }
        grant_headers = {'Content-Type': 'application/x-www-form-urlencoded'}

        async with session.post(self.token_uri, data=grant_data, headers=grant_headers) as resp:
            if resp.status != 200:
                raise Exception(f'Grant failed: {resp.status}')
            result = await resp.json()
            user = OAuth2User(result)

        self.user = user


from dataclasses import asdict
import sys, json, asyncio

from .webapi import *
from .oauth1 import OAuth1
from .oauth2 import OAuth2
from .asyncy import end_loop_workaround

end_loop_workaround()

args = sys.argv[1:]

match args:
    case ['oauth1-flow', app_file, user_file]:
        app = OAuth1(json.load(open(app_file, 'r')))
        asyncio.run(
            app.user_auth_flow('127.0.0.1', 8080))
        with open(user_file, 'w') as f:
            json.dump(asdict(app.user), f, indent=4)
    case ['oauth2-flow', app_file, user_file, *scopes]:
        app = OAuth2(json.load(open(app_file, 'r')))
        scopes = ' '.join(scopes)
        asyncio.run(
            app.user_auth_flow('localhost', 8080, scopes=scopes))
        assert(app.user is not None)
        with open(user_file, 'w') as f:
            json.dump(app.user.to_dict(), f, indent=4)
    case ['scaffold', kind, app_file]:
        if kind == 'oauth1':
            example = {
                'key': '',
                'secret': '',
                'request_uri': '',
                'authorize_uri': '',
                'access_uri': ''
            }
        elif kind == 'oauth2':
            example = {
                'id': '',
                'secret': '',
                'token_uri': '',
                'auth_uri': ''
            }
        else:
            raise ValueError(f"Unknown kind: {kind}")
        with open(app_file, 'w') as f:
            json.dump(example, f, indent=4)
    case _:
        print("Usage:")
        print("  SlyAPI <command> [<args>]")
        print("")
        print("Commands:")
        print("  oauth1-flow APP_FILE USER_FILE [--host HOST] [--port PORT]: grant a single OAuth1 user token with the local flow.")
        print("")
        print("  oauth2-flow APP_FILE USER_FILE SCOPES... [--host HOST] [--port PORT] : grant a single OAuth2 user token with the local flow.")
        print("")
        print("  scaffold KIND APP_FILE: create an example application json file for filling in manually.")
        print("                  KIND is one of 'oauth1' or 'oauth2'.")
        print("")
        print("  help: this dialog.")


from dataclasses import dataclass
import weakref
from copy import deepcopy
from enum import Enum
from typing import Any, AsyncGenerator, Generic, TypeVar, cast

from aiohttp import ClientSession, ClientResponse
from aiohttp.client_exceptions import ContentTypeError

from .asyncy import end_loop_workaround, AsyncInit, AsyncLazy
from .auth import Auth
from .web import Request, Method

Json = dict[str, Any]

class APIError(Exception):
    status: int
    reason: Any

    def __init__(self, status: int, reason: Any):
        super().__init__()
        self.status = status
        self.reason = reason

    def __str__(self) -> str:
        return super().__str__() + F"\nStatus: {self.status}\nReason: {self.reason}"


async def api_err(response: ClientResponse, result: Any = None) -> APIError:
    match result:
        case {'message': msg}:
            return APIError(response.status, msg)
        case _:
            return APIError(response.status, await response.text())


class _EnumParams:
    '''
        Emulate an EnumParam for serialization into URL params.
        Separate class and hidden since Enum's have special behavior.
    '''
    params: dict[str, set[str]]

    def __init__(self):
        self.params = {}

    def __add__(self, other: 'EnumParam|_EnumParams') -> 'EnumParam':
        new_instance = deepcopy(self)
        match other:
            case EnumParam():
                other_items = [(other.get_title(), {other.value})]
            case _EnumParams():
                other_items = other.params.items()
            case _:
                raise
        for k, v in other_items:
            if k not in new_instance.params:
                new_instance.params[k] = set()
            new_instance.params[k] |= v
        return cast(EnumParam, new_instance)

    def to_dict(self, delimiter: str = ',') -> dict[str, str]:
        '''
            Convert packed parameters to a dictionary for use in a URL.
        '''
        return {
            title: delimiter.join(values)
            for title, values in self.params.items()
        }


Self = TypeVar('Self')


class EnumParam(Enum):
    '''
        Collection of API url parameters which have only specific values.
        Serializes to a dictionary for use in a URL.
    '''

    def get_title(self) -> str:
        return self.__class__.__name__[0].lower() + self.__class__.__name__[1:]

    def __add__(self: Self, other: 'EnumParam|_EnumParams') -> Self:
        '''Collect with another parameter or set of parameters.'''
        # return type is compatible with EnumParam for + and to_dict
        return _EnumParams() + self + other  # type: ignore

    def to_dict(self, _delimiter: str = ',') -> dict[str, str]:
        '''
            Convert packed parameters to a dictionary for use in a URL.
        '''
        return {
            self.get_title(): self.value
        }


def convert_url_params(p: Json | None) -> dict[str, str]:
    '''Excludes empty-valued parameters'''
    if p is None: return {}
    return {k: str(v) for k, v in p.items() if v is not None and v != ''}

T = TypeVar('T')
@dataclass
class APIObj(Generic[T]):
    _service: T

    def __init__(self, service: T):
        self._service = service

class WebAPI(AsyncInit):
    base_url: str
    session: ClientSession
    auth: Auth | None

    def __init__(self, auth: Auth | None = None):
        
        self.auth = auth

    async def _async_init(self):
        self._async_ready = True
        # the aiohttp context manager does no asynchronous work when entering.
        # Using the context is not necessary as long as ClientSession.close() 
        # is called.
        self.session = await ClientSession().__aenter__()

        # Although ClientSession.close() may be a coroutine, but it is not
        # necessary to await it as of the time of writing, since it's 
        # connector objects delegate their own coroutine close() methods to 
        # only synchronous methods.
        # this method ensures that the session is closed when the WebAPI object
        # is garbage collected.
        # noinspection PyProtectedMember
        self._finalize = weakref.finalize(self, self.session._connector._close)  # type: ignore ## reportPrivateUsage

        end_loop_workaround()

    def close(self):
        '''Closes the http session with the API server. Should be automatic.'''
        self._finalize()

    def get_full_url(self, path: str) -> str:
        return self.base_url + path

    async def _req_json(self, req: Request) -> Any:
        async with req.send(self.session) as resp:
            try:
                result = await resp.json()
            except ContentTypeError:
                result = await resp.text()
            if resp.status != 200:
                raise await api_err(resp, result)
            return result

    async def _req_empty(self, req: Request) -> None:
        async with req.send(self.session) as resp:
            if resp.status != 204:
                raise await api_err(resp)

    async def _call(self, method: Method, path: str, params: Json | None,
        json: Any, data: Any, headers: dict[str, str] | None = None
        ) -> dict[str, Any]:
        full_url = self.get_full_url(path)
        if json is not None:
            data_ = json
            data_is_json = True
        else:
            data_ = data
            data_is_json = False
        if data_ is None:
            data_: dict[str, Any] = {}
        if headers is None:
            headers = {}
        req = Request(method, full_url, convert_url_params(params), headers, data_, data_is_json)
        if self.auth is not None:
            req = await self.auth.sign_request(self.session, req)
        return await self._req_json(req)

    async def get_json(self, path: str, params: Json | None=None, 
        json: Any=None, data: Any=None, headers: dict[str, str] | None=None
        ) -> dict[str, Any]:
        return await self._call(Method.GET, path, params, json, data, headers)

    async def post_json(self, path: str, params: Json | None=None, 
        json: Any=None, data: Any=None, headers: dict[str, str] | None=None
        ) -> dict[str, Any]:
        return await self._call(Method.POST, path, params, json, data, headers)

    async def put_json(self, path: str, params: Json | None=None, 
        json: Any=None, data: Any=None, headers: dict[str, str] | None=None
        ) -> dict[str, Any]:
        return await self._call(Method.PUT, path, params, json, data, headers)

    @AsyncLazy.wrap
    # TODO: google only?
    async def paginated(self,
                        path: str,
                        params: Json,  # non-const
                        limit: int | None) -> AsyncGenerator[Any, None]:
        '''Return an awaitable and async iterable over google-style paginated items'''
        result_count = 0

        while True:
            page = await self.get_json(path, params)

            items = page.get('items')

            if not items: break

            # result_count += len(items)
            for item in items:
                result_count += 1
                yield item
                if limit is not None and result_count >= limit:
                    return

            page_token = cast(str, page.get('nextPageToken'))
            if not page_token: break
            params['pageToken'] = page_token

# Endpoint = Callable[[WebAPI, T], Coro[U]]

# def decorator(func: Endpoint[Any, T]) -> Endpoint[Any, T]:
#     async def wrapper(self: WebAPI, params: Json) -> T:
#         return await func(self, params)
#     return wrapper



from abc import ABC, abstractmethod
import functools
from typing import Coroutine, ParamSpec, TypeVar, Callable, Generator, Generic, AsyncGenerator, Any

T = TypeVar('T')
U = TypeVar('U')

T_Params = ParamSpec("T_Params")
U_Params = ParamSpec("U_Params")

def end_loop_workaround():
    # workaround for:
    # https://github.com/aio-libs/aiohttp/issues/4324#issuecomment-733884349

    # Replace the event loop destructor thing with one a wrapper which ignores
    # this specific exception on windows.
    import sys
    if sys.platform.startswith("win"):
        # noinspection PyProtectedMember
        from asyncio.proactor_events import _ProactorBasePipeTransport  # type: ignore

        base_del = _ProactorBasePipeTransport.__del__
        if not hasattr(base_del, '_once'):
            def quiet_delete(*args, **kwargs):  # type: ignore
                try:
                    return base_del(*args, **kwargs)  # type: ignore
                except RuntimeError as e:
                    if str(e) != 'Event loop is closed':
                        raise

            quiet_delete._once = True  # type: ignore

            _ProactorBasePipeTransport.__del__ = quiet_delete


TSelfAtAsyncClass = TypeVar("TSelfAtAsyncClass", bound="AsyncInit")


class AsyncInit(ABC): # Awaitable[TSelfAtAsyncClass]
    '''
    Class which depends on some asynchronous initialization.
    To use, override _async_init() to do the actual initialization.
    Accessing any non-static public attributes before the async initialization is complete will result in an error.
    '''
    _async_ready = False
    _async_init_coro: Coroutine[Any, Any, Any] | None = None

    # implementation must initialize the instance
    # arguments should be already passed to the constructor
    # and this method must set self._async_ready to True
    @abstractmethod
    async def _async_init(self): pass

    # not passing args from constructor to _async_init() anymore
    # both are expected to be implemented in the subclass
    # def __init__(self, *args: Any, **kwargs: Any):
    #     if hasattr(self, '_async_init'):
    #         self._async_init_coro = getattr(self, '_async_init')(*args, **kwargs)
    #     else:
    #         raise RuntimeError("AsyncInit class must implement _async_init().")

    def __await__(self: TSelfAtAsyncClass) -> Generator[Any, Any, TSelfAtAsyncClass]:
        async def combined_init() -> TSelfAtAsyncClass:
            # if self._async_init_coro is None:
            #     raise RuntimeError("Expected AsyncInit subclass to set an initialization coroutine.")
            # else:
            await self._async_init()
            return self
        return combined_init().__await__()

# protect members from being accessed before async initialization is complete
def _AsyncInit_get_attr(self: AsyncInit, name: str) -> Any:
    # name not in ('_async_init_coro', '_async_ready', '__await__')
    # private attributes are allowed to be accessed
    # TODO: consider if all private attributes should be allowed
    # note: order of checks is important
    if not name.startswith('_') and not self._async_ready: # type: ignore
        raise RuntimeError("AsyncInit class must be awaited before accessing public attributes.")
    else:
        return object.__getattribute__(self, name)
# workaround to preserve type checks for accessing undefined methods
# Pylance, at least, will presume __getattribute__ will succeed and
# not raise an AttributeError
setattr(AsyncInit, '__getattribute__', _AsyncInit_get_attr)


class AsyncLazy(Generic[T]):
    '''Does not accumulate any results unless awaited.'''
    gen: AsyncGenerator[T, None]

    def __init__(self, gen: AsyncGenerator[T, None]):
        self.gen = gen

    def __aiter__(self) -> AsyncGenerator[T, None]:
        return self.gen

    async def _items(self) -> list[T]:
        return [t async for t in self.gen]

    def __await__(self) -> Generator[Any, None, list[T]]:
        return self._items().__await__()

    def map(self, f: Callable[[T], U]) -> 'AsyncTrans[U]':
        return AsyncTrans(self.gen, f)

    @classmethod
    def wrap(cls, fn: Callable[T_Params, AsyncGenerator[T, None]]):
        @functools.wraps(fn)
        def wrapped(*args: T_Params.args, **kwargs: T_Params.kwargs) -> AsyncLazy[T]:
            return AsyncLazy(fn(*args, **kwargs))
        return wrapped


class AsyncTrans(Generic[U], AsyncLazy[Any]):
    '''
    Does not accumulate any results unless awaited.
    Transforms the results of the generator using the mapping function.
    '''
    gen: AsyncGenerator[Any, None]
    mapping: Callable[[Any], U]

    def __init__(self, gen: AsyncGenerator[Any, None], mapping: Callable[[Any], U]):
        super().__init__(gen)
        self.mapping = mapping

    def __aiter__(self):
        return (self.mapping(t) async for t in self.gen)

    def __await__(self) -> Generator[Any, None, list[U]]:
        return self._items().__await__()

    async def _items(self) -> list[U]:
        return [u async for u in self]


from SlyAPI import *

class TestAPI(WebAPI):

    def __init__(self):
        super().__init__()
        self.attr_initialized_after_await = None

    async def _async_init(self):
        await super()._async_init()
        self.attr_initialized_after_await = 1

async def test_error_uninitialized():

    api_unawaited = TestAPI()

    error = None

    try:
        print( api_unawaited.attr_initialized_after_await )
    except RuntimeError as e:
        error = e

    assert error is not None

    api_awaited = await api_unawaited

    print( api_awaited.attr_initialized_after_await )


import asyncio
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, cast

import aiohttp
import aiohttp.web

class Method(Enum):
    GET = 'GET'
    POST = 'POST'
    PUT = 'PUT'
    DELETE = 'DELETE'
    PATCH = 'PATCH'

@dataclass
class Request:
    method: Method
    url: str
    query_params: dict[str, Any] = field(default_factory=dict)
    headers: dict[str, Any] = field(default_factory=dict)
    data: dict[str, Any] = field(default_factory=dict)
    data_is_json: bool = False

    def send(self, session: aiohttp.ClientSession):
        json = None
        data = None
        params = None
        headers = None
        if self.data_is_json:
            json = self.data
        elif self.data:
            data = self.data
        if self.headers:
            headers = self.headers
        if self.query_params:
            params = self.query_params
        return session.request(self.method.value, self.url, json=json, data=data, params=params, headers=headers)

    def __str__(self) -> str:
        return F"{self.method.value} {self.url} {self.query_params} {self.headers} {self.data}"

async def serve_one_request(host: str, port: int, response_body: str) -> dict[str, str]:

    query: dict[str, str] = {}
    handled_req = False

    server = aiohttp.web.Application()

    async def index_handler(request: aiohttp.web.Request):
        nonlocal query, handled_req
        query = cast(dict[str, str], request.query)
        handled_req = True
        return aiohttp.web.Response(text=response_body, content_type='text/html')
    server.router.add_get("/", index_handler)

    run_task_ = aiohttp.web._run_app(server, host=host, port=port) # type: ignore ## reportPrivateUsage
    run_task = asyncio.create_task(run_task_)

    while not handled_req:
        await asyncio.sleep(0.1)
    run_task.cancel()
    try:
        await run_task
    except asyncio.exceptions.CancelledError:
        pass

    return query


import base64, hmac, secrets
from datetime import datetime
from hashlib import sha1
from typing import Any

from dataclasses import dataclass

import aiohttp

from .auth import Auth
from .web import Method, Request, serve_one_request

# https://datatracker.ietf.org/doc/html/rfc5849


# https://datatracker.ietf.org/doc/html/rfc5849#section-3.6
def percentEncode(s: str):
    result = ''
    URLSAFE = b'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-._~'
    for c in s.encode('utf8'):
        result += chr(c) if c in URLSAFE else f'%{c:02X}'
    return result

# https://datatracker.ietf.org/doc/html/rfc5849#section-3.4.1.3.2
def paramString(params: dict[str, Any]) -> str:
    results: list[str] = []
    encoded = {
        percentEncode(k): percentEncode(v) for k, v in params.items()
    }
    for k, v in sorted(encoded.items()):
        results.append(f'{k}={v}')
    return '&'.join(results)

# https://datatracker.ietf.org/doc/html/rfc5849#section-3.4.2
def _hmac_sign(request: Request, signing_params: dict[str, Any], appSecret: str, userSecret: str|None = None) -> str:
        all_params = request.data | request.query_params | signing_params
        base = F"{request.method.value.upper()}&{percentEncode(request.url.lower())}&{percentEncode(paramString(all_params))}"
        # NOTE:
        #  2.  An "&" character (ASCII code 38), which MUST be included
        #        even when either secret is empty.
        signingKey = percentEncode(appSecret) + '&'
        if userSecret is not None:
            signingKey +=  percentEncode(userSecret)

        hashed = hmac.new( bytes(signingKey,'ascii'),bytes(base, 'ascii'), sha1).digest() #.rstrip(b'\n')
        return base64.b64encode(hashed).decode('ascii')

def _common_oauth_params(appKey: str):
    nonce = base64.b64encode(secrets.token_bytes(32)).strip(b'+/=').decode('ascii')
    timestamp = str(int(datetime.utcnow().timestamp()))
    # nonce = base64.b64encode(b'a'*32).strip(b'+/=').decode('ascii')
    # timestamp = '1'
    return {
        'oauth_consumer_key': appKey,
        'oauth_nonce': nonce,
        'oauth_signature_method': 'HMAC-SHA1',
        'oauth_timestamp': timestamp,
        'oauth_version': '1.0'
    }

@dataclass
class OAuth1User:
    key: str
    secret: str

    def __init__(self, source: dict[str, Any]):
        match source:
            case { 'key': key, 'secret': secret }: # dumps
                self.key = key
                self.secret = secret
            case {
                'oauth_token': key, 
                'oauth_token_secret': secret,
                **_others
            }: # OAuth1 grant
                self.key = key
                self.secret = secret
            case _:
                raise ValueError(F"Invalid OAuth1User source: {source}")

class OAuth1(Auth):
    key: str
    secret: str

    request_uri: str # flow step 1
    authorize_uri: str # flow step 2
    access_uri: str # flow step 3

    user: OAuth1User | None = None

    def __init__(self, source: dict[str, Any], user: OAuth1User | None = None) -> None:
        match source:
            case { 'key': key, 'secret': secret, 'request_uri': request_uri, 'authorize_uri': authorize_uri, 'access_uri': access_uri }: # dumps
                self.key = key
                self.secret = secret
                self.request_uri = request_uri
                self.authorize_uri = authorize_uri
                self.access_uri = access_uri
            case _:
                raise ValueError(F"Invalid OAuth1 source: {source}")
        self.user = user

    async def refresh(self, session: aiohttp.ClientSession):
        if self.user is None:
            raise NotImplemented("Client credentials are not refreshable")
        raise NotImplemented("Refresh is not implemented")

    async def sign_request(self, session: aiohttp.ClientSession, request: Request, do_user: bool = True) -> Request:
        signing_params = _common_oauth_params(self.key)
        user_secret = None
        if self.user is not None and do_user:
            signing_params['oauth_token'] = self.user.key
            user_secret = self.user.secret

        signature = _hmac_sign(request, signing_params, self.secret, user_secret)

        oauth_params = signing_params | { 'oauth_signature': signature }
        oauth_params_str = ', '.join(F'{percentEncode(k)}="{percentEncode(v)}"' for k, v in sorted(oauth_params.items()))
        oauth_headers =  {
            'Authorization': F"OAuth {oauth_params_str}",
        }
        
        request.headers |= oauth_headers

        return request

    async def user_auth_flow(self, redirect_host: str, redirect_port: int, **kwargs: str):
        import webbrowser
        import urllib.parse

        redirect_uri = F'http://{redirect_host}:{redirect_port}'

        session = aiohttp.ClientSession()

        # step 1: get a token to ask the user for authorization
        request = Request(
            Method.POST,
            self.request_uri,
            {'oauth_callback': percentEncode(redirect_uri)}
        )
        signed_request = await self.sign_request(session, request, do_user=False)
        
        async with signed_request.send(session) as resp:
            content = await resp.text()
            resp_params = urllib.parse.parse_qs(content)
            oauth_token = resp_params['oauth_token'][0]
            # oauth_token_secret = resp_params['oauth_token_secret'][0]
            if resp_params['oauth_callback_confirmed'][0] != 'true':
                raise ValueError(f"oauth_callback_confirmed was not true")

        # step 2: get the user to authorize the application
        grant_link = F"{self.authorize_uri}?{urllib.parse.urlencode({'oauth_token': oauth_token})}"

        webbrowser.open(grant_link, new=1, autoraise=True)

        # step 2 (cont.): wait for the user to be redirected with the code
        query = await serve_one_request(redirect_host, redirect_port, '<html><body>You can close this window now.</body></html>')

        oauth_token = query['oauth_token']
        oauth_verifier = query['oauth_verifier']

        # step 3: exchange the code for access token
        # this step does not use the OAuth authorization headers
        async with session.request('POST', self.access_uri, params = {
            'oauth_token': oauth_token,
            'oauth_verifier': oauth_verifier
        }) as resp:
            content = await resp.text()
            resp_params = urllib.parse.parse_qs(content)
            self.user = OAuth1User({k: v[0] for k, v in resp_params.items()})



from abc import ABC, abstractmethod
from typing import Any, TypeVar
import json

import aiohttp

from .web import Request

TSelf = TypeVar('TSelf')

class Auth(ABC):

    @abstractmethod
    async def refresh(self, session: aiohttp.ClientSession): pass

    @abstractmethod
    async def sign_request(self, session: aiohttp.ClientSession, request: Request, do_user: bool = True) -> Request: pass

    @abstractmethod
    async def user_auth_flow(self, redirect_host: str, redirect_port: int, **kwargs: str): pass


class APIKey(Auth):
    params: dict[str, str]

    def __init__(self, param_name: str, secret: str):
        self.params = {param_name: secret}

    @classmethod
    def from_file(cls, path: str):
        with open(path) as f:
            data = json.load(f)
        if len(data.keys()) != 1:
            raise ValueError('Unknown API Key format. Should be JSON: {ParamName: Key}')
        return cls(*data.items().pop())

    async def refresh(self, *args: Any, **kwargs: Any): raise NotImplemented()

    async def sign_request(self, session: aiohttp.ClientSession, request: Request, do_user: bool = True) -> Request:
        request.query_params |= self.params
        return request

    async def user_auth_flow(self, *args: Any, **kwargs: Any): raise NotImplemented()

    # @classmethod
    # async def flow(cls: type[TSelf], **kwargs: str) -> TSelf:
        
    #     print('Please retrieve an API key from the service provider.')
    #     input("Press Enter to continue...")
        
    #     destination_file = kwargs.get('destination')
    #     if destination_file is None:
    #         destination_file = input('Enter destination file path: ')

    #     api_key_param = kwargs.get('param_name')
    #     if api_key_param is None:
    #         api_key_param = input('Enter query parameter name: ')

    #     api_key_secret = kwargs.get('secret')
    #     if api_key_secret is None:
    #         api_key_secret = input('Enter API key secret: ')

    #     obj = cls(api_key_param, api_key_secret)

    #     with open(destination_file, 'w') as f:
    #         json.dump(obj.to_dict(), f)

    #     print(f'API key saved to {destination_file}')

    #     return obj


